unit NLib;

interface

uses NTypes;

type
  { Exceptions }

  Exception = class(TObject)
    Msg: string;
    constructor Create(const Msg: string);
  end;

  EIOError = class(Exception)
    ErrNo: Integer;
    Path: string;
    constructor Create(const Path: string);
  end;

  EOSError = class(Exception)
    ErrNo: Integer;
    constructor Create;
  end;

  { Types and Classes }

  TCmpInt = function (const A, B: Integer): Integer;
  TCmpStr = function (const A, B: string): Integer;

  TSeekWhence = (swSet, swCur, swEnd);

  TStream = class(TObject)
  protected
    ReadP, EndP, Buf: PChar;
  public
    destructor Destroy; override;
    function Read(var Buf; NBytes: Cardinal): Cardinal; virtual; abstract;
    function Write(const Buf; NBytes: Cardinal): Cardinal; virtual; abstract;
    function Seek(Offset: Int64; Whence: TSeekWhence=swSet): Int64; virtual;
        abstract;
    function Seekable: Boolean; virtual; abstract;
    function Size: Cardinal; virtual; abstract;
    function ReadLine(var S: string): Boolean;
    function ReadLines: TStringArray;
  end;
  
  TBytesStream = class(TStream)
    Data: PChar;
    DataSize: Cardinal;
    Position: Cardinal;
    HasData: Boolean;
    constructor Create(const InitialBytes; NBytes: Cardinal;
        NoCopy: Boolean=False); overload;
    constructor Create(const S: string; NoCopy: Boolean=False); overload;
    destructor Destroy; override; 
    function Read(var Buf; NBytes: Cardinal): Cardinal; override;
    function ReadNoCopy(var Buf: PChar; NBytes: Cardinal): Cardinal;
    function Write(const Buf; NBytes: Cardinal): Cardinal; override;
    function Seek(Offset: Int64; Whence: TSeekWhence = swSet): Int64; override;
    function Seekable: Boolean; override;
    function Size: Cardinal; override;
  end;

function DivMod(A, B: Integer; var R: Integer): Integer;

function SplitCommandLine(const S: string): TStringArray;
function JoinCommandLine(const Args: array of string): string;

function IStr(Value: Longint): string;
function I64Str(Value: Int64): string;
function FStr(Value: Double; Digits: Integer): string;
function IHex(Value: Longint): string;
function I64Hex(Value: Int64): string;
function UnitFormat(Value: Int64): string;
function NumFormat(Value: Int64): string;

function SInt(const S: string): Longint;
function SInt64(const S: string): Int64;
function SFloat(const S: string): Double;
function HInt(const S: string): Longint;

function FindBytes(const Buf; NBytes: Integer; const Sub: string;
    Start: Integer=0): Integer;

function SliceArray(const StrArray: array of string; Start: Integer):
    TStringArray;
function StringArray(const Args: array of string): TStringArray;

function BaseName(const Path: string): string;
function DirName(const Path: string): string;

function RoundOff(Value: Double): Longint;

function Cmp(const A, B: Integer): Integer; overload;
function Cmp(const A, B: string): Integer; overload;
procedure Sort(var IntArray: array of Integer; Cmp: TCmpInt=nil); overload;
procedure Sort(var StrArray: array of string; Cmp: TCmpStr=nil); overload;

procedure Print(const S: string = ''); overload;
procedure Print(const S: string; const Eol: string); overload;
procedure Print(const F: TStream; const S: string = ''); overload;
procedure Print(const F: TStream; const S: string; const Eol: string); overload;
function FormatException(const E: TObject): string;
procedure SysFatal;

implementation

uses NSys, NStr;

function DivMod(A, B: Integer; var R: Integer): Integer;
begin
  // R := A mod B;
  // Result := A div B;
  asm
    mov eax, A
    cdq
    idiv B
    mov @Result, eax
    mov eax, R
    mov [eax], edx
  end;
end;

function SplitCommandLine(const S: string): TStringArray;
var
  i, argc: Integer;
  q, qq: Boolean;
  t: string;
begin
  SetLength(Result, 1024);
  argc := 0;
  q := False;
  qq := False;
  i := 0;
  t := '';
  while i < Length(S) do
  begin
    Inc(i);
    if qq then
    begin
      t := t + S[i];
      qq := False;
      Continue;
    end;
    case S[i] of
      '"':
        if q and (S[i+1] = '"') then
        begin
          qq := True;
          Continue;
        end else
          q := not q;
      ' ', #9, #13, #10:
        if q then
          t := t + S[i]
        else if t <> '' then
        begin
          Result[argc] := t;
          if argc = High(Result) then
            SetLength(Result, Length(Result) + 1024);
          Inc(argc);
          t := '';
        end;
      else
        t := t + S[i]
    end;
  end;
  if t <> '' then
  begin
    Result[argc] := t;
    Inc(argc);
  end;
  SetLength(Result, argc);
end;

function JoinCommandLine(const Args: array of string): string;
  function InSpaceQuotes(const S: string): Boolean;
  var
    i, q: Integer;
  begin
    q := -1;
    for i := 1 to Length(S) do
      case S[i] of
        #9, ' ':
          begin
            Result := True;
            Exit;
          end;
        '"':
          if i = q+1 then
          begin
            Result := True;
            Exit;
          end else
            q := i;
      end;
    Result := False;
  end;
var
  i, j, len: Integer;
  sp: Boolean;
  s: string;
begin
  len := 0;
  for i := 0 to High(Args) do
    Inc(len, Length(Args[i])*2+3); // +3: quote*2+space, *2: '"' -> '""'
  SetString(Result, nil, len);
  j := 1;
  for i := 0 to High(Args) do
  begin
    s := Args[i];
    s := Replace(s, '"', '""');
    sp := InSpaceQuotes(s);
    if sp then
    begin
      Result[j] := '"';
      Inc(j);
    end;
    Move(s[1], Result[j], Length(s));
    Inc(j, Length(s));
    if sp then
    begin
      Result[j] := '"';
      Inc(j);
    end;
    Result[j] := ' ';
    Inc(j);
  end;
  SetLength(Result, j-2); // strip last space
end;

function IStr(Value: Longint): string;
var
  sign: Boolean;
  m, i: Integer;
  buf: array[0..19] of Char;
begin
  if Value < 0 then
  begin
    sign := True;
    Value := -Value;
  end else
    sign := False;
  i := SizeOf(buf);
  repeat
    Dec(i);
    // m := Value mod 10;
    // Value := Value div 10;
    Value := DivMod(Value, 10, m);
    buf[i] := Char(m + Ord('0'));
  until Value = 0;
  if sign then
  begin
    Dec(i);
    buf[i] := '-';
  end;
  SetString(Result, PChar(@buf[i]), SizeOf(buf)-i);
end;

function I64Str(Value: Int64): string;
var
  sign: Boolean;
  m, i: Integer;
  buf: array[0..31] of Char;
begin
  if Value < 0 then
  begin
    sign := True;
    Value := -Value;
  end else
    sign := False;
  i := SizeOf(buf);
  repeat
    Dec(i);
    m := Value mod 10;
    Value := Value div 10;
    buf[i] := Char(m + Ord('0'));
  until Value = 0;
  if sign then
  begin
    Dec(i);
    buf[i] := '-';
  end;
  SetString(Result, PChar(@buf[i]), SizeOf(buf)-i);
end;

function Pow10(N: Integer): Double;
const
  tab: array[0..159] of Double = (
    1e0,  1e1,  1e2,  1e3,  1e4,  1e5,  1e6,  1e7,  1e8,  1e9,
    1e10, 1e11, 1e12, 1e13, 1e14, 1e15, 1e16, 1e17, 1e18, 1e19,
    1e20, 1e21, 1e22, 1e23, 1e24, 1e25, 1e26, 1e27, 1e28, 1e29,
    1e30, 1e31, 1e32, 1e33, 1e34, 1e35, 1e36, 1e37, 1e38, 1e39,
    1e40, 1e41, 1e42, 1e43, 1e44, 1e45, 1e46, 1e47, 1e48, 1e49,
    1e50, 1e51, 1e52, 1e53, 1e54, 1e55, 1e56, 1e57, 1e58, 1e59,
    1e60, 1e61, 1e62, 1e63, 1e64, 1e65, 1e66, 1e67, 1e68, 1e69,
    1e70, 1e71, 1e72, 1e73, 1e74, 1e75, 1e76, 1e77, 1e78, 1e79,
    1e80, 1e81, 1e82, 1e83, 1e84, 1e85, 1e86, 1e87, 1e88, 1e89,
    1e90, 1e91, 1e92, 1e93, 1e94, 1e95, 1e96, 1e97, 1e98, 1e99,
    1e100,1e101,1e102,1e103,1e104,1e105,1e106,1e107,1e108,1e109,
    1e110,1e111,1e112,1e113,1e114,1e115,1e116,1e117,1e118,1e119,
    1e120,1e121,1e122,1e123,1e124,1e125,1e126,1e127,1e128,1e129,
    1e130,1e131,1e132,1e133,1e134,1e135,1e136,1e137,1e138,1e139,
    1e140,1e141,1e142,1e143,1e144,1e145,1e146,1e147,1e148,1e149,
    1e150,1e151,1e152,1e153,1e154,1e155,1e156,1e157,1e158,1e159
  );
var
  m: Integer;
begin
  if N < 0 then
  begin
    N := -N;
    if N < Length(tab) then
      Result := 1 / tab[N]
    else begin
      m := n div 2;
      Result := 1 / (Pow10(m) * Pow10(N-m));
    end;
  end
  else begin
    if N < Length(tab) then
      Result := tab[N]
    else begin
      m := n div 2;
      Result := Pow10(m) * Pow10(N-m);
    end;
  end;
end;

function FStr(Value: Double; Digits: Integer): string;
const
  rlog2 = 0.30102999566398114; // 1/log2(10)
var
  m: Double;
  i, ndig, k, e2, e10, l, len: Integer;
  p0, p1: PInteger;
begin
  p0 := @Value;
  p1 := p0;
  Inc(p1); // little endian

  e2 := p1^ shr 20 and $7ff; // biased exponent
  if e2 = $7ff then
  begin
    if (p0^ = 0) then
    begin
      if (p1^ and $800FFFFF = 0) then
        Result := 'Inf'
      else if (p1^ and $800FFFFF = $80000000) then
        Result := '-Inf'
    end else
      Result := 'NaN';
    Exit;
  end;

  if Value = 0.0 then
  begin
    if Digits = 0 then
      Result := '0'
    else
      Result := '0.' + StringOfChar('0', Digits);
    Exit;
  end;

  Dec(e2, 1022);
  {
    Value = 2^e2 * f
    Value = 10^e10 * m
    e10 = e2 * 1/log2(10)
    m = Value * 1/10^e10
  }
  e10 := Trunc(e2 * rlog2);
  m := Value * Pow10(-e10);

  len := Digits + 3; // +3: '-' and '0.'
  if e10 > 0 then
    Inc(len, e10);
  SetString(Result, nil, len);
  k := 1;
  if m < 0 then
  begin
    m := -m;
    Result[k] := '-';
    Inc(k);
  end;

  if m <> 0 then
  begin
    while m < 1.0 do
    begin
      m := m * 10.0;
      Dec(e10);
    end;
    while m > 10.0 do
    begin
      m := m / 10.0;
      Inc(e10);
    end;
  end;

  m := m + Pow10(-e10-Digits-1); // round off

  ndig := 0;
  if e10 < 0 then
  begin
    Result[k] := '0';
    Inc(k);
    if Digits = 0 then
    begin
      SetLength(Result, k-1);
      Exit;
    end;

    Result[k] := '.';
    Inc(k);

    for i := 1 to -e10-1 do
    begin
      if ndig = Digits then
      begin
        SetLength(Result, k-1);
        Exit;
      end;
      Result[k] := '0';
      Inc(k);
      Inc(ndig);
    end;
  end;

  l := 0;
  while True do
  begin
    i := Trunc(m);
    if not ((i = 0) and (l = 0) and (e10 > 0)) then
    begin
      Result[k] := Char(i+Ord('0'));
      Inc(k);
    end;
    if l > e10 then
    begin
      Inc(ndig);
      if ndig = Digits then
        Break;
    end;
    if l = e10 then
    begin
      if Digits = 0 then
        Break;
      Result[k] := '.';
      Inc(k);
    end;
    Inc(l);
    m := (m - i) * 10;
  end;
  SetLength(Result, k-1);
end;

function IHex(Value: Longint): string;
const
  hexdigs: array[0..15] of Char = '0123456789abcdef';
var
  sign: Boolean;
  i: Integer;
  buf: array[0..9] of Char;
begin
  if Value < 0 then
  begin
    sign := True;
    Value := -Value;
  end else
    sign := False;
  i := SizeOf(buf);
  repeat
    Dec(i);
    buf[i] := hexdigs[Value and $F];
    Value := Value shr 4;
  until Value = 0;
  if sign then
  begin
    Dec(i);
    buf[i] := '-';
  end;
  SetString(Result, PChar(@buf[i]), SizeOf(buf)-i);
end;

function I64Hex(Value: Int64): string;
const
  hexdigs: array[0..15] of Char = '0123456789abcdef';
var
  sign: Boolean;
  i: Integer;
  buf: array[0..17] of Char;
begin
  if Value < 0 then
  begin
    sign := True;
    Value := -Value;
  end else
    sign := False;
  i := SizeOf(buf);
  repeat
    Dec(i);
    buf[i] := hexdigs[Value and $F];
    Value := Value shr 4;
  until Value = 0;
  if sign then
  begin
    Dec(i);
    buf[i] := '-';
  end;
  SetString(Result, PChar(@buf[i]), SizeOf(buf)-i);
end;

// convert to 4 digits or 3 digits + point and unit character
function UnitFormat(Value: Int64): string;
const
  KiroSize = 1024;
  MegaSize = 1024 * 1024;
  GigaSize = 1024 * 1024 * 1024;
var
  u: Char;
  d: Integer;
  n: Double;
  sign: Boolean;
begin
  if Value < 0 then
  begin
    sign := True;
    Value := -Value;
  end else
    sign := False;
  if Value < KiroSize then
  begin
    if sign then
      Value := -Value;
    Result := IStr(Value)
  end
  else begin
    if Value >= MegaSize * 1000 then
    begin
      n := Value / GigaSize;
      u := 'G';
    end
    else if Value >= KiroSize * 1000 then
    begin
      n := Value / MegaSize;
      u := 'M';
    end
    else begin
      n := Value / KiroSize;
      u := 'K';
    end;

    // truncate n
    if n >= 100 then
    begin
      d := 0;
      n := Int(n);
    end
    else if n >= 10 then
    begin
      d := 1;
      n := Int(n * 10) / 10;
    end
    else begin
      d := 2;
      n := Int(n * 100) / 100;
    end;
    if sign then
      n := -n;
    Result := FStr(n, d) + u;
  end;
end;

function NumFormat(Value: Int64): string;
const
  n: Integer = 3;
var
  s: string;
  i, j, len, ofs: Integer;
begin
  s := I64Str(Value);
  len := Length(s);
  SetLength(Result, len + (len-1) div n);
  ofs := n - len mod n;
  i := 1;
  for j := 1 to len do
  begin
    Result[i] := s[j];
    Inc(i);
    if (j + ofs) mod 3 = 0 then
    begin
      Result[i] := ',';
      Inc(i);
    end;
  end;
end;

function SInt(const S: string): Longint;
var
  sign: Boolean;
  i: Integer;
begin
  sign := False;
  i := 1;
  while (i <= Length(S)) and (S[i] in [#9, ' ']) do
    Inc(i);
  if i <= Length(S) then
    if S[i] = '+' then
      Inc(i)
    else if S[i] = '-' then
    begin
      sign := True;
      Inc(i);
    end;
  Result := 0;
  while (i <= Length(S)) and (S[i] in ['0'..'9']) do
  begin
    Result := Result * 10 + Ord(S[i]) - Ord('0');
    Inc(i);
  end;
  if sign then
    Result := -Result;
end;

function SInt64(const S: string): Int64;
var
  sign: Boolean;
  i: Integer;
begin
  sign := False;
  i := 1;
  while (i <= Length(S)) and (S[i] in [#9, ' ']) do
     Inc(i);
  if (i <= Length(S)) and (S[i] = '-') then
    if S[i] = '+' then
      Inc(i)
    else if S[i] = '-' then
    begin
      sign := True;
      Inc(i);
    end;
  Result := 0;
  while (i <= Length(S)) and (S[i] in ['0'..'9']) do
  begin
    Result := Result * 10 + Ord(S[i]) - Ord('0');
    Inc(i);
  end;
  if sign then
    Result := -Result;
end;

// TODO: do not use Val()
function SFloat(const S: string): Double;
var
  code: Integer;
begin
  Val(S, Result, code);
end;

function HInt(const S: string): Longint;
var
  sign: Boolean;
  i: Integer;
begin
  sign := False;
  i := 1;
  while (i <= Length(S)) and (S[i] in [#9, ' ']) do
    Inc(i);
  if i <= Length(S) then
    if S[i] = '+' then
      Inc(i)
    else if S[i] = '-' then
    begin
      sign := True;
      Inc(i);
    end;
  Result := 0;
  while i <= Length(S) do
  begin
    case S[i] of
      '0'..'9': Result := Result * 16 + Ord(S[i]) - Ord('0');
      'A'..'F': Result := Result * 16 + Ord(S[i]) - 55;
      'a'..'f': Result := Result * 16 + Ord(S[i]) - 87;
    else
      Break;
    end;
    Inc(i);
  end;
  if sign then
    Result := -Result;
end;


function FindBytes(const Buf; NBytes: Integer; const Sub: string;
    Start: Integer=0): Integer;
var
  i, j, endi, sublen: Integer;
  p: PChar;
begin
  Result := -1;
  sublen := Length(Sub);
  if sublen = 0 then
    Exit;
  endi := NBytes - sublen + 1;
  if endi < Start then
    Exit;
  p := @Buf;
  for i := Start to endi do
    if p[i] = Sub[1] then
    begin
      j := 1;
      while (j < sublen) and (p[i+j] = Sub[1+j]) do 
        Inc(j);
      if j = sublen then
      begin
        Result := i;
        Exit;
      end;
    end;
end;

function SliceArray(const StrArray: array of string; Start: Integer):
    TStringArray;
var
  i, len: Integer;  
begin
  len := High(StrArray) - Start + 1;
  if len < 0 then
    len := 0;
  SetLength(Result, len);
  for i := 0 to High(Result) do
    Result[i] := StrArray[Start+i];
end;

function StringArray(const Args: array of string): TStringArray;
var
  i: Integer;
begin
  SetLength(Result, Length(Args));
  for i := 0 to High(Args) do
    Result[i] := Args[i];
end;

function BaseName(const Path: string): string;
var
  s: string;
begin
  SplitPath(Path, s, Result);
end;

function DirName(const Path: string): string;
var
  s: string;
begin
  SplitPath(Path, Result, s);
end;

function RoundOff(Value: Double): Longint;
begin
  if Value >= 0 then
    Result := Trunc(Value + 0.5)
  else
    Result := Trunc(Value - 0.5);
end;

function Cmp(const A, B: Integer): Integer;
begin
  if A > B then
    Result := 1
  else if A < B then
    Result := -1
  else
    Result := 0;
end;

function Cmp(const A, B: string): Integer;
begin
  if A > B then
    Result := 1
  else if A < B then
    Result := -1
  else
    Result := 0;
end;

procedure Sort(var IntArray: array of Integer; Cmp: TCmpInt=nil);
  procedure merge(var Src, Dest: array of Integer; L, H: Integer);
  var
    i, j, k, m: Integer;
  begin
    if L >= H then
      Exit;
    m := (H + L) div 2;
    merge(Dest, Src, L, m);
    merge(Dest, Src, m+1, H);
    j := L;
    k := m+1;
    for i := L to H do
    begin
      if (k > H) or (j <= m) and (Cmp(Src[j], Src[k]) <= 0) then
      begin
        Dest[i] := Src[j];
        Inc(j);
      end
      else begin 
        Dest[i] := Src[k];
        Inc(k);
      end;
    end;
  end;
var
  i: Integer;
  aux: array of Integer;
begin
  if not Assigned(Cmp) then
    Cmp := NLib.Cmp;
  SetLength(aux, Length(IntArray));
  for i := 0 to High(aux) do
    aux[i] := IntArray[i];
  merge(aux, IntArray, 0, High(aux));
end;

procedure Sort(var StrArray: array of string; Cmp: TCmpStr=nil);
  procedure merge(var Src, Dest: array of string; L, H: Integer);
  var
    i, j, k, m: Integer;
  begin
    if L >= H then
      Exit;
    m := (H + L) div 2;
    merge(Dest, Src, L, m);
    merge(Dest, Src, m+1, H);
    j := L;
    k := m+1;
    for i := L to H do
    begin
      if (k > H) or (j <= m) and (Cmp(Src[j], Src[k]) <= 0) then
      begin
        Dest[i] := Src[j];
        Inc(j);
      end
      else begin 
        Dest[i] := Src[k];
        Inc(k);
      end;
    end;
  end;
var
  i: Integer;
  aux: array of string;
begin
  if not Assigned(Cmp) then
    Cmp := NLib.Cmp;
  SetLength(aux, Length(StrArray));
  for i := 0 to High(aux) do
    aux[i] := StrArray[i];
  merge(aux, StrArray, 0, High(aux));
end;

procedure Print(const S: string = '');
var
  t: string;
begin
  t := S + EOL;
  StdOut.Write(t[1], Length(t));
end;

procedure Print(const S: string; const Eol: string);
var
  t: string;
begin
  t := S + Eol;
  StdOut.Write(t[1], Length(t));
end;

procedure Print(const F: TStream; const S: string = '');
var
  t: string;
begin
  t := S + EOL;
  F.Write(t[1], Length(t));
end;

procedure Print(const F: TStream; const S: string; const Eol: string);
var
  t: string;
begin
  t := S + Eol;
  F.Write(t[1], Length(t));
end;

function FormatException(const E: TObject): string;
begin
  Result := BaseName(ProgName) + ': ';
  if E is EIOError then
    with E as EIOError do
      Result := Result + Path + ': ' + Msg
  else if E is Exception then
    with E as Exception do
      Result := Result + Msg
  else
    Result := 'unknown error';
end;

procedure SysFatal;
begin
  Print(StdErr, FormatException(ExceptObject));
  Halt(1);
end;


{ Exceptions }

constructor Exception.Create(const Msg: string);
begin
  Self.Msg := Msg;
end;

constructor EIOError.Create(const Path: string);
begin
  ErrNo := SysErrNo;
  Msg := SysError(ErrNo);
  Self.Path := Path;
end;

constructor EOSError.Create;
begin
  ErrNo := SysErrNo;
  Msg := SysError(ErrNo);
end;


{ TStream }

destructor TStream.Destroy;
begin
  if Buf <> nil then
    FreeMem(Buf);
  inherited;
end;

function TStream.ReadLine(var S: string): Boolean;
const
  bufsize = 8192;
var
  len, dlen: Integer;
  sp: PChar;
  foundlf: Boolean;
begin
  if Buf = nil then
    GetMem(Buf, bufsize);

  dlen := 0;
  foundlf := False;
  while not foundlf do
  begin
    if ReadP = EndP then
    begin
      len := Read(Buf^, bufsize);
      if len = 0 then
      begin
        Result := dlen <> 0;
        Exit;
      end;
      ReadP := Buf;
      EndP := Buf + len;
    end;

    sp := ReadP;
    while (ReadP < EndP) and (not foundlf) do
    begin
      foundlf := ReadP^ = #10;
      Inc(ReadP);
    end;

    SetLength(S, dlen + (ReadP - sp));
    Move(sp^, S[dlen+1], ReadP - sp);
    Inc(dlen, ReadP - sp);
  end;
  Result := True;
end;

function TStream.ReadLines: TStringArray;
var
  i: Integer;
  s: string;
begin
  SetLength(Result, 1024);
  i := 0;
  while ReadLine(s) do
  begin
    if i = Length(Result) then
      SetLength(Result, i + i div 4);
    Result[i] := s;
    Inc(i);
  end;
  SetLength(Result, i);
end;

{ TBytesStream }

constructor TBytesStream.Create(const InitialBytes; NBytes: Cardinal;
    NoCopy: Boolean=False);
begin
  DataSize := NBytes;
  if NoCopy then
    Data := @InitialBytes
  else begin
    // TODO: raise exception if out of memory
    GetMem(Data, DataSize);
    HasData := True;
    Move(InitialBytes, Data^, DataSize);
  end;
  Seek(0);
end;

constructor TBytesStream.Create(const S: string; NoCopy: Boolean=False);
begin
  Create(S[1], Length(S), NoCopy);
end;

destructor TBytesStream.Destroy;
begin
  if HasData then
    FreeMem(Data);
end;

function TBytesStream.Read(var Buf; NBytes: Cardinal): Cardinal;
begin
  if NBytes > (DataSize - Position) then
    NBytes := DataSize - Position;
  Move(Data[Position], Buf, NBytes);
  Inc(Position, NBytes);
  Result := NBytes;
end;

function TBytesStream.ReadNoCopy(var Buf: PChar; NBytes: Cardinal): Cardinal;
begin
  if NBytes > (DataSize - Position) then
    NBytes := DataSize - Position;
  Buf := @Data[Position];
  Inc(Position, NBytes);
  Result := NBytes;
end;
  
function TBytesStream.Write(const Buf; NBytes: Cardinal): Cardinal;
begin
  if NBytes > (DataSize - Position) then
    NBytes := DataSize - Position;
  Move(Buf, Data[Position], NBytes);
  Inc(Position, NBytes);
  Result := NBytes;
end;

// TODO: support swCur and swEnd
function TBytesStream.Seek(Offset: Int64; Whence: TSeekWhence = swSet): Int64;
begin
  Position := Offset;
  if Position > DataSize then
    Position := DataSize;
  Result := Position;
end;

function TBytesStream.Seekable: Boolean;
begin
  Result := True;
end;

function TBytesStream.Size: Cardinal;
begin
  Result := DataSize;
end;

end.
